package net.sergeych.kiloparsec.adapter

import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.channels.SendChannel
import kotlinx.coroutines.flow.Flow
import kotlinx.coroutines.flow.flow
import kotlinx.coroutines.suspendCancellableCoroutine
import kotlinx.coroutines.withContext
import net.sergeych.kiloparsec.Transport
import java.net.InetAddress
import java.net.InetSocketAddress
import java.nio.channels.AsynchronousServerSocketChannel
import java.nio.channels.AsynchronousSocketChannel
import kotlin.coroutines.suspendCoroutine

actual fun NetworkAddress(host: String, port: Int): NetworkAddress =
    JvmNetworkAddress(InetAddress.getByName(host), port)

actual fun acceptTcpDevice(port: Int): Flow<Transport.Device> {
    return flow {
        println("start generating tcp accept flow")
        val socket = withContext(Dispatchers.IO) {
            AsynchronousServerSocketChannel.open().also {
                it.bind(InetSocketAddress(InetAddress.getLocalHost(), port))
            }
        }
        while (true) {
            println("Server socket ready $socket")
            val connectedSocket = suspendCancellableCoroutine { continuation ->
                continuation.invokeOnCancellation {
                    socket.close()
                }
                socket.accept(continuation, ContinuationHandler())
            }
            println("incoming connection")
            emit(asyncSocketToDevice(connectedSocket))
        }
    }
}

actual suspend fun connectTcpDevice(address: NetworkAddress): Transport.Device {
    address as JvmNetworkAddress
    val socket = withContext(Dispatchers.IO) {
        AsynchronousSocketChannel.open()
    }
    suspendCoroutine { cont ->
        socket.connect(address.socketAddress, cont, VoidCompletionHandler)
    }
    println("connected")
    return asyncSocketToDevice(socket)
}

suspend fun SendChannel<UByte>.sendAll(bytes: Collection<UByte>) {
    for (b in bytes) send(b)
}